import asyncio
import heapq
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any

from .base import Backend, BackendError


@dataclass(order=True)
class ScheduledTask:
    scheduled_time: datetime
    task_id: str = field(compare=False)


class MemoryBackend(Backend):

    def __init__(self) -> None:
        self._task_metadata: dict[str, dict[str, dict[str, Any]]] = defaultdict(dict)  # {QUEUE_NAME: {TASK_ID: task_data}}
        self._pending: dict[str, list[str]] = defaultdict(list)
        self._scheduled: dict[str, list[ScheduledTask]] = defaultdict(list)
        self._crons: dict[str, dict[str, dict[str, Any]]] = defaultdict(dict)

        self._locks: dict[str, asyncio.Lock] = defaultdict(asyncio.Lock)  # for thread-safety
        self._connected = False

    async def connect(self) -> None:
        self._connected = True

    async def disconnect(self) -> None:
        self._connected = False

    @property
    def is_connected(self) -> bool:
        return self._connected

    async def _create_tasks(self, queue_name: str, tasks: list[dict[str, Any]]) -> list[bool]:
        self._check_connected()

        async with self._locks[queue_name]:
            success = []
            for task in tasks:
                if task["id"] not in self._task_metadata[queue_name]:
                    self._task_metadata[queue_name][task["id"]] = task
                    success.append(True)
                else:
                    success.append(False)

            return success

    async def append(self, queue_name: str, tasks: list[dict[str, Any]], unique: bool = True) -> list[bool]:
        self._check_connected()

        if unique:
            success = await self._create_tasks(queue_name, tasks)
            to_queue = [t for i, t in enumerate(tasks) if success[i]]
        else:
            success = [True] * len(tasks)
            to_queue = tasks

        async with self._locks[queue_name]:
            for task in to_queue:
                if not unique:
                    self._task_metadata[queue_name][task["id"]] = task

                self._pending[queue_name].append(task["id"])

            return success

    async def pop(self, queue_name: str, limit: int = 1, timeout: float | None = None) -> list[dict[str, Any]]:
        self._check_connected()

        start_time = asyncio.get_event_loop().time()

        while True:
            async with self._locks[queue_name]:
                if self._pending[queue_name]:
                    tasks = []
                    q = self._pending[queue_name]

                    for _ in range(min(limit, len(q))):
                        task_id = q.pop(0)
                        task_data = self._task_metadata[queue_name].get(task_id)
                        if task_data:
                            tasks.append(task_data)

                    return tasks

            if timeout is None or timeout <= 0:
                return []

            elapsed = asyncio.get_event_loop().time() - start_time
            if elapsed >= timeout:
                return []

            await asyncio.sleep(min(0.05, timeout - elapsed))

    async def get_pending(self, queue_name: str, count: int = 1) -> list[dict[str, Any]]:
        self._check_connected()

        async with self._locks[queue_name]:
            task_ids = list(self._pending[queue_name])[:count]

            tasks = []
            for t in task_ids:
                if task_data := self._task_metadata[queue_name].get(t):
                    tasks.append(task_data)

            return tasks


    async def size(self, queue_name: str) -> int:
        self._check_connected()

        async with self._locks[queue_name]:
            return len(self._pending[queue_name])

    async def clear(self, queue_name: str) -> int:
        self._check_connected()

        async with self._locks[queue_name]:
            queue_size = len(self._task_metadata[queue_name])
            queue_cron_size = len(self._crons[queue_name])

            self._task_metadata[queue_name].clear()
            self._pending[queue_name].clear()
            self._scheduled[queue_name].clear()
            self._crons[queue_name].clear()

            return queue_size + queue_cron_size

    async def get_tasks(self, queue_name: str, task_ids: list[str]) -> dict[str,dict[str, Any]]:
        self._check_connected()

        async with self._locks[queue_name]:
            results = {}
            for task_id in task_ids:
                result = self._task_metadata[queue_name].get(task_id)
                if result:
                    results[task_id] = result

            return results

    async def schedule(self, queue_name: str, task_data: dict[str, Any], at: datetime, unique: bool = True) -> bool:
        self._check_connected()

        if unique:
            success = await self._create_tasks(queue_name, [task_data])
            if not success[0]:
                return False

        async with self._locks[queue_name]:
            if not unique:
                self._task_metadata[queue_name][task_data["id"]] = task_data

            scheduled_task = ScheduledTask(at, task_data["id"])
            heapq.heappush(self._scheduled[queue_name], scheduled_task)

            return True

    async def pop_scheduled(self, queue_name: str, now: datetime | None = None) -> list[dict[str, Any]]:
        self._check_connected()

        if now is None:
            now = datetime.now(timezone.utc)

        async with self._locks[queue_name]:
            tasks = []
            scheduled_tasks = self._scheduled[queue_name]

            while scheduled_tasks and scheduled_tasks[0].scheduled_time <= now:
                scheduled_task = heapq.heappop(scheduled_tasks)
                task_data = self._task_metadata[queue_name].get(scheduled_task.task_id)
                if task_data:
                    tasks.append(task_data)

            return tasks

    async def store_result(self, queue_name: str, task_data: dict[str, Any]) -> bool:
        self._check_connected()

        async with self._locks[queue_name]:
            self._task_metadata[queue_name][task_data['id']] = task_data

            return True

    async def get_results(self, queue_name: str, task_ids: list[str], timeout: float | None = None) -> dict[str,dict[str, Any]]:
        self._check_connected()

        start_time = asyncio.get_event_loop().time()

        if not task_ids:
            return {}

        results = {}
        remaining_ids = task_ids[:]

        while True:
            async with self._locks[queue_name]:
                for task_id in task_ids:
                    task_data = self._task_metadata[queue_name].get(task_id, {})

                    if task_data.get("finished_at"):
                        results[task_id] = task_data
                        remaining_ids.remove(task_id)

            if not remaining_ids:
                return results

            if timeout is None or timeout < 0:
                return results

            # endless wait if timeout == 0
            if timeout == 0:
                await asyncio.sleep(0.05)
                continue

            elapsed = asyncio.get_event_loop().time() - start_time
            if elapsed >= timeout:
                raise TimeoutError(f"Did not complete within {timeout} seconds")

            await asyncio.sleep(min(0.05, timeout - elapsed))

    async def get_stats(self, queue_name: str) -> dict[str, int]:
        self._check_connected()

        async with self._locks[queue_name]:
            return {
                "pending": len(self._pending[queue_name]),
                "completed": len([t for t in self._task_metadata[queue_name].values() if t["finished_at"]]),
                "scheduled": len(self._scheduled[queue_name]),
            }

    async def get_all_tasks(self, queue_name: str) -> list[dict[str, Any]]:
        self._check_connected()

        async with self._locks[queue_name]:
            tasks = self._task_metadata[queue_name]
            return list(tasks.values())

    async def list_queues(self) -> dict[str, int]:
        self._check_connected()

        queues = {}
        for queue_name in self._task_metadata:
            async with self._locks[queue_name]:
                queues[queue_name] = len(self._pending[queue_name])

        return queues

    async def get_scheduled(self, queue_name: str) -> list[dict[str, Any]]:
        self._check_connected()

        async with self._locks[queue_name]:
            tasks = []
            for scheduled_task in self._scheduled[queue_name]:
                task_data = self._task_metadata[queue_name].get(scheduled_task.task_id)
                if task_data:
                    tasks.append(task_data)

            return tasks

    async def add_cron(self, queue_name: str, deterministic_id: str, task_cron: dict[str, Any]) -> bool:
        self._check_connected()

        async with self._locks[queue_name]:
            if deterministic_id not in self._crons[queue_name]:
                self._crons[queue_name][deterministic_id] = task_cron
                return True
            return False

    async def delete_cron(self, queue_name: str, deterministic_id: str) -> bool:
        self._check_connected()

        async with self._locks[queue_name]:
            if deterministic_id in self._crons[queue_name]:
                del self._crons[queue_name][deterministic_id]
                return True
            return False

    async def get_crons(self, queue_name: str) -> list[dict[str, Any]]:
        self._check_connected()

        async with self._locks[queue_name]:
            return list(self._crons[queue_name].values())

    def _check_connected(self) -> None:
        if not self.is_connected:
            raise BackendError("Not connected")



import asyncio
import logging
import signal
from functools import partial

from pydantic import BaseModel

from .backend.base import Backend
from .models import Task, TaskCron
from .queue import Queue
from .utils.task_execution import (
    TaskProcessor,
    generate_unique_worker_id,
)

logger = logging.getLogger(__name__)


class WorkerStats(BaseModel):
    processed: int = 0
    failed: int = 0


WORKER_PREFIX = "<Worker> "
SCHEDULER_PREFIX = "<Scheduler> "
CRON_MANAGER_PREFIX = "<CronManager> "


class Worker:
    def __init__(
        self,
        queue_name: str | list[str],
        backend: Backend,
        shutdown_timeout: float = 30.0,
        max_concurrent_tasks: int = 10,
        enable_job_processing: bool = True,
        enable_scheduler: bool = True,
        enable_cron_manager: bool = True,
    ):
        if not any([enable_job_processing, enable_scheduler, enable_cron_manager]):
            raise ValueError("At least one processing type must be enabled")

        self._backend = backend

        if not isinstance(queue_name, list|tuple):
            queue_name = [str(queue_name)]
        self.queues = [Queue(backend, q) for q in queue_name]

        self.shutdown_timeout = shutdown_timeout
        self.worker_id = generate_unique_worker_id("worker")
        self.stats = WorkerStats()

        self._task_processor = TaskProcessor()
        self._task_semaphore = asyncio.Semaphore(max_concurrent_tasks)
        self._shutdown_event = asyncio.Event()
        self._ctrl_c_counter = 0

        self._blocking_timeout = 5
        self._scheduler_polling_interval = 1.0
        self._cron_polling_interval = 10.0

        self._active_tasks: dict[str, dict[asyncio.Task[Task], Task]] = {queue.name: {} for queue in self.queues}

        self.enable_job_processing = enable_job_processing
        self.enable_scheduler = enable_scheduler
        self.enable_cron_manager = enable_cron_manager

        self._work_queue_tasks: list[asyncio.Task[None]] = []
        self._scheduler_task: asyncio.Task[None] | None = None
        self._cron_manager_task: asyncio.Task[None] | None = None

        self._tasks_to_process: int | None = None
        self._empty_queues: list[str] = []

    async def work(self, max_tasks: int | None = None, oneshot: bool = False) -> None:
        # register signals
        loop = asyncio.get_event_loop()
        self.__register_signal_handlers(loop)

        self._tasks_to_process = max_tasks
        self._empty_queues.clear()

        # reset state (likely relevant only for tests)
        self._shutdown_event.clear()
        self._ctrl_c_counter = 0

        # test connection
        await self._verify_connection(self._backend)

        # start scheduler
        if self.enable_scheduler:
            self._scheduler_task = asyncio.create_task(self._run_scheduler(self._scheduler_polling_interval))

        # start cron manager
        if self.enable_cron_manager:
            self._cron_manager_task = asyncio.create_task(self._run_cron_manager(self._cron_polling_interval))

        # start job processing
        if self.enable_job_processing:
            for queue in self.queues:
                self._work_queue_tasks.append(asyncio.create_task(self._run_worker_loop(queue, oneshot)))

        # blocking wait for created asyncio tasks
        _futures = self._work_queue_tasks
        _futures += [self._scheduler_task] if self._scheduler_task else []
        _futures += [self._cron_manager_task] if self._cron_manager_task else []
        await asyncio.gather(*_futures, return_exceptions=True)
        self._shutdown_event.set()

        # this is starting to feel like Perl
        remaining_tasks = {k: v for inner_dict in self._active_tasks.values() for k, v in inner_dict.items()}

        # attempt to exit cleanly
        if remaining_tasks:
            logger.info(WORKER_PREFIX + f"Waiting for {len(remaining_tasks)} active tasks to complete...")
            try:
                await asyncio.wait_for(
                    asyncio.gather(*remaining_tasks.keys(), return_exceptions=True),
                    timeout=self.shutdown_timeout
                )
            except asyncio.TimeoutError:
                logger.warning("Some tasks did not complete within shutdown timeout")

                # ! FIXME - what should we do here with the existing tasks? (maybe DLQ?)

                for task_future in remaining_tasks:
                    if not task_future.done():
                        task_future.cancel()

                        # ! FIXME - should we try reqeueue here or just store state?
                        # task = remaining_tasks[task_future]
                        # try:
                        #     await queue.add(task)
                        # except Exception as e:
                        #     logger.error(f"Failed to requeue task {task.id}: {e}", exc_info=True)

        # unregister signals
        for sig in (signal.SIGTERM, signal.SIGINT):
            loop.remove_signal_handler(sig)

        logger.info(f"Worker stopped. Processed: {self.stats.processed}, Failed: {self.stats.failed}")

    async def _run_scheduler(self, poll_interval: float) -> None:
        logger.info(SCHEDULER_PREFIX + "started")

        while not self._shutdown_event.is_set():
            try:
                for queue in self.queues:
                    tasks = await queue.enqueue_scheduled()

                    if tasks:
                        _l = len(tasks)
                        _task_s = ", ".join([str(task.id) for task in tasks])
                        logger.info(SCHEDULER_PREFIX + f"Enqueued {_l} scheduled task{'s' if _l > 1 else ''} for processing: {_task_s}")

            except asyncio.CancelledError:
                logger.warning(SCHEDULER_PREFIX + "cancelled")
                break

            except Exception as e:
                logger.exception(SCHEDULER_PREFIX + f"Scheduling failed with error: {e}", exc_info=True)
                self._shutdown_event.set()
                break

            await asyncio.sleep(poll_interval)  # TODO: replace polling with notifications when worker notifications are implemented

        logger.info(SCHEDULER_PREFIX + "stopped")

    async def _run_cron_manager(self, poll_interval: float) -> None:
        logger.info(CRON_MANAGER_PREFIX + "started")

        while not self._shutdown_event.is_set():
            try:
                for queue in self.queues:
                    for cron_data in await queue.get_crons():
                        cron = TaskCron.model_validate(cron_data)

                        _next_run = None
                        for _ in range(3):
                            _next_run = cron.next_run(_next_run)
                            task = cron.create_task(_next_run)
                            success = await queue.schedule(task, at=_next_run)
                            if success:
                                logger.info(CRON_MANAGER_PREFIX + f"Cron {cron.id} ({cron.spec.func}) scheduled at {_next_run}")

            except asyncio.CancelledError:
                logger.warning(CRON_MANAGER_PREFIX + "cancelled")
                break

            except Exception as e:
                logger.exception(CRON_MANAGER_PREFIX + f"failed with error: {e}", exc_info=True)
                self._shutdown_event.set()
                break

            await asyncio.sleep(poll_interval)  # TODO: replace polling with notifications when worker notifications are implemented

        logger.info(CRON_MANAGER_PREFIX + "stopped")


    async def _run_worker_loop(self, queue: Queue, oneshot: bool = False) -> None:
        while not self._shutdown_event.is_set():

            if self._tasks_to_process is not None and self._tasks_to_process <= 0:
                self._shutdown_event.set()
                break

            # clean up completed tasks
            completed = [t for t in self._active_tasks[queue.name] if t.done()]
            for t in completed:
                del self._active_tasks[queue.name][t]

            if self._task_semaphore._value == 0:
                # hacky way to wait until there is an available slot
                async with self._task_semaphore:
                    continue

            # how many tasks to get
            capacity = self._task_semaphore._value
            if self._tasks_to_process is not None:
                capacity = min(capacity, self._tasks_to_process)

            try:
                available_tasks = await queue.pop_pending(timeout=self._blocking_timeout,
                                                limit=capacity)

                if oneshot and not available_tasks:
                    logger.info(WORKER_PREFIX + f"Queue '{queue.name}' emptied")
                    self._empty_queues.append(queue.name)
                    if len(self._empty_queues) == len(self.queues):
                        self._shutdown_event.set()
                    break

                for task in available_tasks:
                    logger.info(WORKER_PREFIX + f"Processing task {task.id} ({task.spec.func})")
                    task_future = asyncio.create_task(self.process_task_semaphore_wrap(queue, task))
                    self._active_tasks[queue.name][task_future] = task

                    if self._tasks_to_process is not None:
                        self._tasks_to_process -= 1

            except asyncio.CancelledError:
                logger.warning(WORKER_PREFIX + "cancelled")
                break

            except Exception as e:
                logger.exception(WORKER_PREFIX + f"failed with error: {e}", exc_info=True)
                self._shutdown_event.set()
                break

    async def process_task_semaphore_wrap(self, queue: Queue, task: Task) -> Task:
        async with self._task_semaphore:
            task = await self.process_task(task)
            await self._store_result(queue, task)

            # schedule the task for retry
            if task.error and task.should_retry and task.next_retry_at is not None:
                await queue.retry(task, task.next_retry_at)

            return task

    async def process_task(self, task: Task) -> Task:

        exception, task = await self._task_processor.execute_task(task, self.worker_id)

        if task.completed:
            self.stats.processed += 1
            logger.info(WORKER_PREFIX + f"Task {task.id} completed successfully")
        else:
            self.stats.failed += 1

        # non retriable task
        if task.error and not task.is_retriable:
            logger.error(WORKER_PREFIX + f"Task {task.id} failed: {exception}", exc_info=True)

        # retriable task - final failure
        if task.error and task.is_retriable and not task.should_retry:
            logger.error(WORKER_PREFIX + f"Task {task.id} failed after {task.retry_count} retries: {exception}", exc_info=True)

        # retriable task - reschedule
        if task.error and task.should_retry:
            logger.warning(WORKER_PREFIX + f"Task {task.id} failed (attempt {task.retry_count}/{task.config.retry}), scheduling retry at {task.next_retry_at}")

        return task

    async def _store_result(self, queue: Queue, task: Task) -> None:
        try:
            await queue.backend.store_result(queue.name, task.model_dump(mode='json'))  # TODO
        except Exception:
            logger.exception(f"Failed to store result for task {task.id}", exc_info=True)

    def __register_signal_handlers(self, loop: asyncio.AbstractEventLoop) -> None:
        CTRL_C_THRESHOLD = 3
        def signal_handler(sig: signal.Signals) -> None:
            if self._shutdown_event.is_set():
                if self._ctrl_c_counter == CTRL_C_THRESHOLD:
                    logger.warning("Forcing shutdown...")
                    # cancel all tasks on shutdown
                    _futures = self._work_queue_tasks
                    _futures += [
                        k for inner_dict in self._active_tasks.values()  # type: ignore[misc]
                          for k, v in inner_dict.items()]
                    _futures += [self._scheduler_task] if self._scheduler_task else []
                    _futures += [self._cron_manager_task] if self._cron_manager_task else []
                    for future in _futures:
                        future.cancel()
                    # we cancelled active tasks, so clear all dicts as well
                    for d in self._active_tasks.values():
                        d.clear()
                    return

                logger.info(f"Press CTRL+C {CTRL_C_THRESHOLD - self._ctrl_c_counter} more times to force shutdown...")
                self._ctrl_c_counter += 1

            else:
                logger.info(f"Received signal {sig.name}, initiating graceful shutdown...")
                self._shutdown_event.set()

        for sig in (signal.SIGTERM, signal.SIGINT):
            loop.add_signal_handler(sig, partial(signal_handler, sig))

    async def _verify_connection(self, backend: Backend) -> bool:
        if not backend.is_connected:
            # TODO: implement backend.ping()
            await backend.connect()

        return True



"""
This file contains utility functions meant for internal use only. Expect breaking changes if you use them directly.
"""

import importlib
import inspect
import os
import socket
import sys
from collections.abc import Callable
from datetime import datetime, timedelta, timezone
from typing import Annotated, Any, cast, get_args, get_origin
from uuid import uuid4

import anyio
from pydantic import ConfigDict, PydanticSchemaGenerationError, TypeAdapter

from ..models import Task
from .fastapi import Depends

cache_main_module: str | None = None
cache_signature: dict[Callable[..., Any], inspect.Signature] = {}


class TaskInternal(Task):
    model_config = ConfigDict(frozen=False)

    @staticmethod
    def from_task(task: Task) -> "TaskInternal":
        return TaskInternal.model_validate(task.model_dump(mode="json"))

    def create_task(self) -> Task:
        return Task.model_validate(self.model_dump(mode="json"))


def generate_unique_worker_id(prefix: str) -> str:
    return f"{prefix}-{socket.gethostname()}-{str(uuid4())[:8]}"


class TaskProcessor:

    @staticmethod
    async def _actually_execute_task(task: TaskInternal) -> Any:
        # resolve the function from its string representation
        func = TaskProcessor.resolve_function(task.spec.func)
        args = task.spec.args or []
        kwargs = task.spec.kwargs or {}

        # validate all parameters, inject DI and Task
        final_args, final_kwargs = await TaskProcessor.process_function_parameters(func, args, kwargs, task)

        # async task
        if inspect.iscoroutinefunction(func):
            return await func(*final_args, **final_kwargs)

        # sync task
        return await anyio.to_thread.run_sync(lambda: func(*final_args, **final_kwargs))

    @staticmethod
    async def execute_task(__task: "Task", worker_id: str) -> tuple[Exception | None, "Task"]:
        task = TaskInternal.from_task(__task)

        try:
            task, _generators = await TaskProcessor.process_pre_task_middleware(task)
        except Exception as e:
            raise Exception("Middleware error") from e

        try:
            result = await TaskProcessor._actually_execute_task(task)
            task = TaskProcessor.handle_success_and_update_task_metadata(task, result, worker_id)
            exception = None

        except Exception as e:
            task = await TaskProcessor.handle_failed_task(task, e)
            exception = e  # temporary

        try:
            task = await TaskProcessor.process_post_task_middleware(task, _generators)
        except Exception as e:
            raise Exception("Middleware error") from e

        return exception, task.create_task()

    @staticmethod
    async def process_pre_task_middleware(task: TaskInternal) -> tuple[TaskInternal, list[Any]]:
        if not task.spec.middleware:
            return task, []

        _generators = []

        for middleware_string in task.spec.middleware:
            middleware = TaskProcessor.resolve_function(middleware_string, wrapped=False)
            gen = middleware(task)
            task = next(gen) or task
            _generators.append(gen)

        return task, _generators

    @staticmethod
    async def process_post_task_middleware(task: TaskInternal, _generators: list[Any]) -> TaskInternal:
        if not _generators:
            return task

        for gen in _generators[::-1]:  # post task middleware goes in reverse order
            try:
                task = gen.send(task) or task
            except StopIteration as e:
                task = e.value or task

        return task

    @staticmethod
    def handle_success_and_update_task_metadata(task: TaskInternal, result: Any, worker_id: str) -> TaskInternal:
        task.result = result
        task.completed = True
        task.error = None

        #task.config.worker = worker_id
        task.finished_at = datetime.now(timezone.utc)

        return task

    @staticmethod
    async def handle_failed_task(task: TaskInternal, exception: Exception) -> TaskInternal:
        task.completed = False
        task.error = str(exception)

        if task.is_retriable:
            task = TaskProcessor.handle_retry(task)
        else:
            task.finished_at = datetime.now(timezone.utc)

        return task

    @staticmethod
    def handle_retry(task: TaskInternal) -> TaskInternal:
        if task.should_retry:
            task.retry_count += 1
            task.last_retry_at = datetime.now(timezone.utc)
            task.next_retry_at = datetime.now(timezone.utc) + timedelta(seconds=TaskProcessor.calculate_retry_delay(task))
            task.finished_at = None
        else:
            task.finished_at = datetime.now(timezone.utc)

        return task

    @staticmethod
    def calculate_retry_delay(task: TaskInternal) -> float:
        if isinstance(task.config.retry_delay, float):
            return task.config.retry_delay  # constant delay for all retries
        if isinstance(task.config.retry_delay, list):
            if len(task.config.retry_delay) == 0:
                return 1.0  # empty list defaults to 1 second  # todo - we should probably refuse empty lists as input

            if task.retry_count < len(task.config.retry_delay):
                return float(task.config.retry_delay[task.retry_count])
            else:
                # use last delay value for remaining retries
                return float(task.config.retry_delay[-1])

        # this should never happen if the library is used correctly
        if isinstance(task.config.retry_delay, int):
            return float(task.config.retry_delay)
        # this should never happen if the library is used correctly
        raise ValueError(f"Invalid retry_delay type: {type(task.config.retry_delay).__name__}. Expected float or list[float].")

    @staticmethod
    def resolve_function(func: str, wrapped: bool = True) -> Callable[..., Any]:
        module_name = None
        function_name = None

        try:
            module_name, function_name = func.split(':')
            module = importlib.import_module(module_name)
            fn = getattr(module, function_name)
            result = fn.__wrapped__ if wrapped else fn
            return cast(Callable[..., Any], result)

        except (ValueError, ImportError, AttributeError) as e:
            # edge case where we are trying to resolve a function from __main__ and worker is running from main
            global cache_main_module
            if not cache_main_module:
                _main_path = os.path.relpath(sys.argv[0])[:-3]  # this handles "python -m app.main" because with "-m" sys.argv[0] is absolute path
                cache_main_module = _main_path.replace(os.sep, ".")  # replace handles situations when user runs "python app/main.py"

            if module_name and function_name and module_name == cache_main_module and "__main__" in sys.modules:  # noqa: SIM102
                if fn := getattr(sys.modules["__main__"], function_name, None):
                    result = fn.__wrapped__ if wrapped else fn
                    return cast(Callable[..., Any], result)

            raise ValueError(f"Cannot resolve function: {func}") from e

    @staticmethod
    async def process_function_parameters(
        func: Callable[..., Any],
        args: list[Any],
        kwargs: dict[str, Any],
        task: TaskInternal | None = None,
    ) -> tuple[list[Any], dict[str, Any]]:

        signature = cache_signature.get(func)
        if not signature:
            signature = inspect.signature(func)
            cache_signature[func] = signature

        final_args = []
        final_kwargs = kwargs.copy()
        remaining_args = args.copy()

        for param_name, param in list(signature.parameters.items()):
            # Task injection (self: Task)
            if task and TaskProcessor._is_task_injection(param):
                final_args.append(task)
                continue

            # validate positional args
            if remaining_args:
                final_args.append(TaskProcessor._validate(remaining_args.pop(0), param.annotation))
                continue

            # dependency injection
            if depends := TaskProcessor.get_depends_from_param(param):
                final_kwargs[param_name] = await TaskProcessor._resolve_dependency(depends.dependency)
                continue

            # validate kwargs
            if param_name in kwargs:
                final_kwargs[param_name] = TaskProcessor._validate(kwargs[param_name], param.annotation)

        return final_args, final_kwargs

    @staticmethod
    def _is_task_injection(param: inspect.Parameter) -> bool:
        if param.name != 'self':
            return False

        if param.annotation == inspect.Parameter.empty:
            return False

        return param.annotation is Task or param.annotation == 'Task'


    @staticmethod
    def _validate(value: Any, annotation: Any) -> Any:
        if value is not None and annotation != inspect.Parameter.empty:
            try:
                return TypeAdapter(annotation).validate_python(value)
            except PydanticSchemaGenerationError:
                pass

        return value

    @staticmethod
    async def _resolve_dependency(func: Callable[..., Any]) -> Any:
        # func = resolver.dependency_overrides.get(dep_func, dep_func)  # ! FIXME

        # resolve nested dependencies
        _, kwargs = await TaskProcessor.process_function_parameters(func, [], {}, None)

        # execute dependency
        if inspect.iscoroutinefunction(func):
            return await func(**kwargs)

        if inspect.isasyncgenfunction(func):
            return await func(**kwargs).__anext__()

        if inspect.isgeneratorfunction(func):
            return await anyio.to_thread.run_sync(next, func(**kwargs))

        return await anyio.to_thread.run_sync(lambda: func(**kwargs))

    @staticmethod
    def get_depends_from_param(param: inspect.Parameter) -> Any | None:
        if param.default != inspect.Parameter.empty and isinstance(param.default, Depends):
            return param.default

        # Annotated style
        if get_origin(param.annotation) is Annotated:
            args = get_args(param.annotation)

            for arg in args[1:]:
                if isinstance(arg, Depends):
                    return arg

        return None



from datetime import datetime, timedelta, timezone
from typing import overload
from uuid import UUID

from .backend.base import Backend
from .models import Task, TaskCron
from .task_factory import TaskFactory


class Queue:

    def __init__(self, backend: Backend, name: str = "default"):
        self.name = name
        self.backend = backend

    @overload
    async def add(self, task: Task) -> bool: ...

    @overload
    async def add(self, task: list[Task]) -> list[bool]: ...

    async def add(self, task: Task | list[Task]) -> bool | list[bool]:
        """Add task into the queue. Accept list of tasks for batch add."""
        await self.__ensure_backend_is_connected()

        if isinstance(task, list):
            batch_mode = True
            tasks = [t.model_dump(mode='json') for t in task]
        else:
            batch_mode = False
            tasks = [task.model_dump(mode='json')]

        success = await self.backend.append(self.name, tasks)

        return success if batch_mode else success[0]

    @overload
    async def get_task(self, task: Task | UUID) -> Task | None: ...

    @overload
    async def get_task(self, task: list[Task | UUID]) -> dict[UUID, Task]: ...

    async def get_task(self, task: Task | UUID | list[Task | UUID]) -> Task | None | dict[UUID, Task]:
        await self.__ensure_backend_is_connected()

        task_ids, batch_mode = self._get_task_ids(task)
        task_results = await self.backend.get_tasks(self.name, task_ids)

        if batch_mode:
            return {UUID(t_id): Task.model_validate(t) for t_id, t in task_results.items()}

        td = task_results.get(task_ids[0])

        return Task.model_validate(td) if td else None

    async def get_all_tasks(self) -> list[Task]:
        """Get all tasks, including completed/failed ones."""
        await self.__ensure_backend_is_connected()
        tasks_data = await self.backend.get_all_tasks(self.name)
        return [Task.model_validate(t) for t in tasks_data]

    async def get_pending(self, count: int = 1) -> list[Task]:
        """List pending tasks."""

        if count <= 0:
            raise ValueError("Value must be larger than zero")

        await self.__ensure_backend_is_connected()

        return [Task.model_validate(t) for t in await self.backend.get_pending(self.name, count)]

    async def schedule(self, task: Task, at: datetime | timedelta) -> bool:
        """Schedule task to be processed after certain time."""
        await self.__ensure_backend_is_connected()

        if isinstance(at, timedelta):
            at = datetime.now(timezone.utc) + at

        if not at.tzinfo:
            raise TypeError("provided datetime must be offset-aware")

        task.__dict__["scheduled_at"] = at

        return await self.backend.schedule(self.name, task.model_dump(mode="json"), at)

    async def get_scheduled(self) -> list[Task]:
        """List scheduled tasks."""
        await self.__ensure_backend_is_connected()
        return [Task.model_validate(t) for t in await self.backend.get_scheduled(self.name)]

    @overload
    async def wait_for(self, task: Task | UUID, timeout: float = 0) -> Task | None: ...

    @overload
    async def wait_for(self, task: list[Task | UUID], timeout: float = 0) -> dict[UUID, Task]: ...

    async def wait_for(self, task: Task | UUID | list[Task | UUID], timeout: float = 0) -> Task | None | dict[UUID, Task]:
        await self.__ensure_backend_is_connected()

        task_ids, batch_mode = self._get_task_ids(task)
        task_results = await self.backend.get_results(self.name, task_ids, timeout)

        if batch_mode:
            return {UUID(t_id): Task.model_validate(t) for t_id, t in task_results.items()}

        td = task_results.get(task_ids[0])

        return Task.model_validate(td) if td else None

    async def retry(self, task: Task | UUID, at: datetime | timedelta | None = None, force: bool = False) -> bool:
        """Retry failed task."""
        _task = await self.get_task(task)  # ensure_backend_is_connected is called in get_task already
        if not _task:
            return False

        if not force and _task.completed:
            raise ValueError("Task has already completed successfully, use force to retry anyways")

        needs_update = False  # temp hack
        if _task.finished_at:
            needs_update = True
            _task.__dict__["last_retry_at"] = datetime.now(timezone.utc)
            _task.__dict__["next_retry_at"] = datetime.now(timezone.utc)
            _task.__dict__["finished_at"] = None

        if at:
            if isinstance(at, timedelta):
                at = datetime.now(timezone.utc) + at

            if not at.tzinfo:
                raise TypeError("provided datetime must be offset-aware")

            if needs_update:
                _task.__dict__["next_retry_at"] = at
                _task.__dict__["scheduled_at"] = at

            return await self.backend.schedule(self.name, _task.model_dump(mode="json"), at, unique=False)

        success = await self.backend.append(self.name, [_task.model_dump(mode="json")], unique=False)
        return success[0]

    async def size(self) -> int:
        """Get number of pending tasks in the queue."""
        await self.__ensure_backend_is_connected()
        return await self.backend.size(self.name)

    async def clear(self) -> int:
        """Clear all tasks, including completed ones."""
        await self.__ensure_backend_is_connected()
        return await self.backend.clear(self.name)

    async def add_cron(self, task: Task, cron: str) -> bool:
        await self.__ensure_backend_is_connected()
        task_cron = TaskFactory.create_cron_from_task(task, cron)
        return await self.backend.add_cron(self.name, str(task_cron.deterministic_id), task_cron.model_dump(mode="json"))

    async def delete_cron(self, task: Task, cron: str) -> bool:
        await self.__ensure_backend_is_connected()
        task_cron = TaskFactory.create_cron_from_task(task, cron)
        return await self.backend.delete_cron(self.name, str(task_cron.deterministic_id))

    async def get_crons(self) -> list[TaskCron]:
        await self.__ensure_backend_is_connected()
        return [TaskCron.model_validate(tc) for tc in await self.backend.get_crons(self.name)]

    async def pop_pending(self, limit: int = 1, timeout: float | None = None) -> list[Task]:
        """Get next task to process. Used internally by workers."""
        await self.__ensure_backend_is_connected()

        if limit <= 0:
            raise ValueError("Pop limit must be greater than zero.")

        tasks_data = await self.backend.pop(self.name, limit, timeout)
        return [Task.model_validate(t) for t in tasks_data]

    async def enqueue_scheduled(self, now: datetime | None = None) -> list[Task]:
        """Enqueue scheduled tasks that are ready to be processed. Used internally by workers."""
        await self.__ensure_backend_is_connected()

        tasks_data = await self.backend.pop_scheduled(self.name, now)
        tasks = [Task.model_validate(t) for t in tasks_data]
        await self.backend.append(self.name, tasks_data, unique=False)

        return tasks

    async def __ensure_backend_is_connected(self) -> None:
        """Automatically connects backend on first async call."""
        if not self.backend.is_connected:
            await self.backend.connect()

    def _get_task_ids(self, task: list[Task | UUID] | Task | UUID) -> tuple[list[str], bool]:
        batch_mode = True
        if not isinstance(task, list):
            task = [task]
            batch_mode = False

        # set to deduplicate task ids
        task_ids = list({str(t.id if isinstance(t, Task) else t) for t in task})

        return task_ids, batch_mode



import importlib
from datetime import datetime, timezone
from typing import (
    Annotated,
    Any,
    ParamSpec,
    TypeVar,
)
from uuid import UUID, uuid4, uuid5

from croniter import croniter
from pydantic import (
    AfterValidator,
    AwareDatetime,
    BaseModel,
    ConfigDict,
    Field,
    TypeAdapter,
    model_validator,
)

P = ParamSpec('P')
R = TypeVar('R')


TASK_CRON_NS = UUID('7005b432-c135-4131-b19e-d3dc89703a9a')


def cron_expression_validator(value: str) -> str:
    if not croniter.is_valid(value):
        raise ValueError(f"{value} is not a valid cron expression")

    return value

CronExpression = Annotated[str, AfterValidator(cron_expression_validator)]


class Spec(BaseModel):
    model_config = ConfigDict(frozen=True)

    func: str
    args: list[Any] = Field(default_factory=list)
    kwargs: dict[str, Any] = Field(default_factory=dict)
    return_type: str | None = None
    middleware: list[str] | None = None


class Config(BaseModel):
    model_config = ConfigDict(frozen=True)

    retry: int = Field(default=0, ge=0)
    retry_delay: float | list[float] = Field(default=1.0)

    # timeout: float | None = None  # seconds
    # tags: dict[str, str] = Field(default_factory=dict)


class Task(BaseModel):
    model_config = ConfigDict(frozen=True)

    id: UUID = Field(default_factory=uuid4)
    completed: bool = False
    error: str | None = None
    result: Any = None

    spec: Spec
    config: Config = Field(default_factory=Config)

    created_at: AwareDatetime = Field(default_factory=lambda: datetime.now(timezone.utc))
    finished_at: AwareDatetime | None = None
    scheduled_at: AwareDatetime | None = None

    retry_count: int = 0
    last_retry_at: AwareDatetime | None = None
    next_retry_at: AwareDatetime | None = None
    # caller: str | None = None
    # worker: str | None = None

    # extra: dict[str, Any] = Field(default_factory=dict)

    @property
    def is_retriable(self) -> bool:
        return self.config.retry > 0

    @property
    def should_retry(self) -> bool:
        return self.config.retry > 0 and self.retry_count < self.config.retry

    @model_validator(mode='after')
    def _reconstruct_pydantic_result(self) -> 'Task':
        """Reconstruct result if it's pydantic model."""

        if self.result and self.spec.return_type:
            # Reconstruct return if it's pydantic model
            module_name, type_name = self.spec.return_type.rsplit('.', 1)
            module = importlib.import_module(module_name)
            return_type = getattr(module, type_name)
            self.__dict__["result"] = TypeAdapter(return_type).validate_python(self.result)

        return self

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        parts = {
            "id": repr(self.id),
            "func": repr(self.spec.func),
            "args": repr(self.spec.args),
            "kwargs": repr(self.spec.kwargs),
            "completed": repr(self.completed),
            "error": repr(self.error)
        }

        if self.retry_count > 0:
            parts["retry_count"] = str(self.retry_count)

        return f"Task({', '.join([f'{k}={v}' for k, v in parts.items()])})"


class TaskCron(BaseModel):
    model_config = ConfigDict(frozen=True)

    id: UUID = Field(default_factory=uuid4)
    expression: CronExpression

    spec: Spec
    config: Config

    # enabled: bool = True
    # last_run: AwareDatetime | None = None
    # next_run: AwareDatetime | None = None

    @property
    def deterministic_id(self) -> UUID:
        """Deterministic UUID to prevent duplicated cron definitions."""
        s = self.spec.model_dump_json() + self.config.model_dump_json() + self.expression
        return uuid5(TASK_CRON_NS, s)

    def next_run(self, start: datetime | None = None) -> datetime:
        if not start:
            start = datetime.now(timezone.utc)
        return croniter(self.expression, start).get_next(datetime)

    def create_task(self, start: datetime) -> Task:
        return Task(
            id=uuid5(TASK_CRON_NS, str(self.deterministic_id) + str(start.timestamp())),
            spec=self.spec.model_copy(deep=True),
            config=self.config.model_copy(deep=True)
        )



import typer
from rich import print as rprint

from sheppy import __version__

from .commands.cron.list import list_crons
from .commands.queue.list import list_queues
from .commands.task.add import add
from .commands.task.info import info
from .commands.task.list import list_tasks
from .commands.task.retry import retry
from .commands.task.schedule import schedule
from .commands.task.test import test
from .commands.work import work

app = typer.Typer(rich_markup_mode="rich", no_args_is_help=True, add_completion=False)


def version_callback(value: bool) -> None:
    if value:
        rprint(f"[green]Sheppy version: {__version__}[/green]")
        raise typer.Exit()


@app.callback()
def callback(
    version: bool = typer.Option(None, "--version", help="Show the version and exit.", callback=version_callback),
) -> None:
    """
    Sheppy - Modern Task Queue
    """
    pass


task_app = typer.Typer(help="Task management commands", no_args_is_help=True)
task_app.command(name="list")(list_tasks)
task_app.command()(info)
task_app.command()(retry)
task_app.command()(test)
task_app.command()(add)
task_app.command()(schedule)

app.add_typer(task_app, name="task")

queue_app = typer.Typer(help="Queue management commands", no_args_is_help=True)
queue_app.command(name="list")(list_queues)

app.add_typer(queue_app, name="queue")

cron_app = typer.Typer(help="Cron management commands", no_args_is_help=True)
cron_app.command(name="list")(list_crons)

app.add_typer(cron_app, name="cron")

app.command()(work)



import asyncio
from datetime import datetime, timedelta, timezone
from typing import overload
from uuid import UUID

from .backend.memory import MemoryBackend
from .models import Task, TaskCron
from .queue import Queue
from .utils.task_execution import TaskProcessor


class TestQueue:
    __test__ = False

    def __init__(
        self,
        name: str = "test-queue",
        #dependency_overrides: dict[Callable[..., Any], Callable[..., Any]] | None = None  # ! FIXME
    ):
        self.name = name

        self._backend = MemoryBackend()
        self._backend._connected = True
        self._queue = Queue(self._backend, self.name)
        #self._dependency_resolver = DependencyResolver(dependency_overrides)
        self._worker_id = "TestQueue"
        self._task_processor = TaskProcessor()

        self.processed_tasks: list[Task] = []
        self.failed_tasks: list[Task] = []

    @overload
    def add(self, task: Task) -> bool: ...

    @overload
    def add(self, task: list[Task]) -> list[bool]: ...

    def add(self, task: Task | list[Task]) -> bool | list[bool]:
        """Add task into the queue. Accept list of tasks for batch add."""
        return asyncio.run(self._queue.add(task))  # type: ignore[return-value]

    @overload
    def get_task(self, task: Task | UUID) -> Task | None: ...

    @overload
    def get_task(self, task: list[Task | UUID]) -> dict[UUID, Task]: ...

    def get_task(self, task: Task | UUID | list[Task | UUID]) -> Task | None | dict[UUID, Task]:
        return asyncio.run(self._queue.get_task(task))

    def get_all_tasks(self) -> list[Task]:
        """Get all tasks, including completed/failed ones."""
        return asyncio.run(self._queue.get_all_tasks())

    def get_pending(self, count: int = 1) -> list[Task]:
        """List pending tasks."""
        return asyncio.run(self._queue.get_pending(count))

    def schedule(self, task: Task, at: datetime | timedelta) -> bool:
        """Schedule task to be processed after certain time."""
        return asyncio.run(self._queue.schedule(task, at))

    def get_scheduled(self) -> list[Task]:
        """List scheduled tasks."""
        return asyncio.run(self._queue.get_scheduled())

    def retry(self, task: Task | UUID, at: datetime | timedelta | None = None, force: bool = False) -> bool:
        """Retry failed task."""
        return asyncio.run(self._queue.retry(task, at, force))

    def size(self) -> int:
        """Get number of pending tasks in the queue."""
        return asyncio.run(self._queue.size())

    def clear(self) -> int:
        """Clear all tasks, including completed ones."""
        return asyncio.run(self._queue.clear())

    def add_cron(self, task: Task, cron: str) -> bool:
        return asyncio.run(self._queue.add_cron(task, cron))

    def delete_cron(self, task: Task, cron: str) -> bool:
        return asyncio.run(self._queue.delete_cron(task, cron))

    def get_crons(self) -> list[TaskCron]:
        return asyncio.run(self._queue.get_crons())

    def process_next(self) -> Task | None:

        async def _process_next_async() -> Task | None:
            tasks = await self._queue.pop_pending(limit=1)
            return await self._execute_task(tasks[0]) if tasks else None

        return asyncio.run(_process_next_async())

    def process_all(self) -> list[Task]:
        processed = []

        while task := self.process_next():
            processed.append(task)

        return processed

    def process_scheduled(self, at: datetime | timedelta | None = None) -> list[Task]:
        if isinstance(at, timedelta):
            at = datetime.now(timezone.utc) + at
        elif at is None:
            at = datetime.now(timezone.utc)

        async def _process_scheduled_async(at: datetime) -> list[Task]:
            tasks = [Task.model_validate(t) for t in await self._backend.pop_scheduled(self.name, at)]
            return [await self._execute_task(task) for task in tasks]

        return asyncio.run(_process_scheduled_async(at))

    async def _execute_task(self, __task: Task) -> Task:
        _, task = await self._task_processor.execute_task(__task, self._worker_id)

        self.processed_tasks.append(task)

        if task.error:
            self.failed_tasks.append(task)

            if task.should_retry:
                # retry immediately
                await self._queue.retry(task)

        await self._backend.store_result(self.name, task.model_dump(mode='json'))

        data = await self._backend.get_tasks(self.name, [str(task.id)])
        stored_task_data = data.get(str(task.id))

        if stored_task_data:
            return Task.model_validate(stored_task_data)

        return task


def assert_is_new(task: Task | None) -> None:
    assert task is not None
    assert isinstance(task, Task)

    assert task.completed is False
    assert task.error is None
    assert task.result is None
    assert task.finished_at is None


def assert_is_completed(task: Task | None) -> None:
    assert task is not None
    assert isinstance(task, Task)

    assert task.completed is True
    assert task.error is None
    assert task.finished_at is not None


def assert_is_failed(task: Task | None) -> None:
    assert task is not None
    assert isinstance(task, Task)

    assert not task.completed
    assert task.error is not None
    assert task.result is None

    if not task.should_retry:
        assert task.finished_at is not None



import logging
from datetime import datetime, timezone
from enum import Enum

from rich.console import Console

from sheppy import Backend, MemoryBackend, RedisBackend

console = Console()


class OutputFormat(str, Enum):
    table = "table"
    json = "json"


class LogLevel(str, Enum):
    """Log level enum for CLI arguments."""
    debug = "debug"
    info = "info"
    warning = "warning"
    error = "error"

    def to_logging_level(self) -> int:
        """Convert to Python logging level."""
        return getattr(logging, self.value.upper())  # type: ignore[no-any-return]


class BackendType(str, Enum):
    """Backend type enum."""
    redis = "redis"


def get_backend(backend_type: BackendType, redis_url: str) -> Backend:
    """Create backend instance based on type."""
    if backend_type == BackendType.redis:
        return RedisBackend(redis_url)

    elif backend_type == BackendType.memory:
        return MemoryBackend()

    else:
        raise ValueError(f"Unknown backend: {backend_type}")


def humanize_datetime(dt: datetime | None, now: datetime | None = None) -> str:
    if not dt:
        return "N/A"

    if not now:
        now = datetime.now(timezone.utc)

    delta = (dt - now).total_seconds()

    is_past = delta < 0
    abs_delta = abs(delta)

    if abs_delta > 86400:  # 24 hours
        return dt.strftime("%Y-%m-%d %H:%M:%S")

    hours = int(abs_delta/3600)
    hours_s = f"{hours} hour" + ("s" if hours > 1 else "")

    minutes = int((abs_delta % 3600) / 60)
    minutes_s = f"{minutes} minute" + ("s" if minutes > 1 else "")

    if abs_delta > 3600:
        time_string = hours_s

    elif abs_delta > 60:
        time_string = minutes_s
    else:
        time_string = f"{int(abs_delta)} second" + ("s" if abs_delta >= 2 else "")

    return f"{time_string} ago" if is_past else f"in {time_string}"



"""
This file contains utility functions meant for internal use only. Expect breaking changes if you use them directly.
"""

__all__ = ["Depends"]

from collections.abc import Callable
from typing import Any

try:
    from fastapi.params import Depends  # type: ignore
except ImportError:
    # fallback implementation if FastAPI is not installed
    class Depends:  # type: ignore[no-redef]
        def __init__(
            self, dependency: Callable[..., Any] | None = None, /
        ):
            self.dependency = dependency



import asyncio
from typing import Annotated

import typer
from rich.table import Table

from sheppy import Queue

from ...utils import BackendType, console, get_backend, humanize_datetime


def list_crons(
    queue: Annotated[str, typer.Option("--queue", "-q", help="Name of queue")] = "default",
    backend: Annotated[BackendType, typer.Option("--backend", "-b", help="Queue backend type")] = BackendType.redis,
    redis_url: Annotated[str, typer.Option("--redis-url", "-r", help="Redis server URL")] = "redis://127.0.0.1:6379",
) -> None:
    """List all active crons."""

    async def _list() -> None:
        backend_instance = get_backend(backend, redis_url)
        q = Queue(backend_instance, queue)

        crons = await q.get_crons()

        if not crons:
            console.print(f"[yellow]No crons found in queue '{queue}'[/yellow]")
            return

        table = Table(title="Active Crons")
        table = Table(title=f"Tasks in [bold cyan]{queue}[/bold cyan] (showing {len(crons)} of {len(crons)})")
        table.add_column("Cron ID", style="dim")
        table.add_column("Function", style="blue")
        table.add_column("Args", style="blue")
        table.add_column("Kwargs", style="blue")
        table.add_column("Cron expression", style="yellow")
        # table.add_column("Last run (UTC)", style="blue")
        table.add_column("Next run (UTC)", style="blue")

        for cron in crons:
            table.add_row(
                str(cron.id),
                str(cron.spec.func),
                str(cron.spec.args),
                str(cron.spec.kwargs),
                str(cron.expression),
                # todo - last run
                humanize_datetime(cron.next_run())
            )

        console.print(table)


    asyncio.run(_list())



import os
import sys
from collections.abc import Callable
from functools import wraps
from typing import (
    Any,
    ParamSpec,
    TypeVar,
    get_type_hints,
    overload,
)

from .models import Config, Spec, Task, TaskCron
from .utils.validation import validate_input

P = ParamSpec('P')
R = TypeVar('R')

cache_main_module: str | None = None
cache_return_type: dict[Callable[..., Any], str | None] = {}


class TaskFactory:

    def __init__(self) -> None:
        pass

    @staticmethod
    def _get_return_type(func: Callable[..., Any]) -> str | None:
        """Get function return type"""
        if func in cache_return_type:
            return cache_return_type.get(func)

        try:
            return_type = get_type_hints(func).get('return')
            if return_type and hasattr(return_type, '__module__') and hasattr(return_type, '__qualname__'):
                return_type = f"{return_type.__module__}.{return_type.__qualname__}"
        except TypeError:
            return_type = None

        cache_return_type[func] = return_type

        return return_type

    @staticmethod
    def _stringify_function(func: Callable[..., Any]) -> str:
        _module = func.__module__
        # special case if the task is in the main python file that is executed
        if _module == "__main__":
            global cache_main_module
            if not cache_main_module:
                # this handles "python -m app.main" because with "-m" sys.argv[0] is absolute path
                _main_path = os.path.relpath(sys.argv[0])[:-3]
                # replace handles situations when user runs "python app/main.py"
                cache_main_module = _main_path.replace(os.sep, ".")

            _module = cache_main_module

        return f"{_module}:{func.__name__}"

    @staticmethod
    def create_task(func: Callable[..., Any],
                    args: list[Any],
                    kwargs: dict[str, Any],
                    retry: float,
                    retry_delay: float | list[float] | None,
                    middleware: list[Callable[..., Any]] | None
                    ) -> Task:
        # Store return type to later reconstruct the result
        return_type = TaskFactory._get_return_type(func)

        task_config: dict[str, Any] = {
            "retry": retry
        }
        if retry_delay is not None:
            task_config["retry_delay"] = retry_delay

        func_string = TaskFactory._stringify_function(func)

        args, kwargs = validate_input(func, list(args or []), dict(kwargs or {}))

        stringified_middlewares = []
        if middleware:
            for m in middleware:
                # todo: should probably also validate them here
                stringified_middlewares.append(TaskFactory._stringify_function(m))

        _task = Task(
            spec=Spec(
                func=func_string,
                args=args,
                kwargs=kwargs,
                return_type=return_type,
                middleware=stringified_middlewares
            ),
            config=Config(**task_config)
        )

        return _task

    @staticmethod
    def create_cron_from_task(task: Task, cron_expression: str) -> TaskCron:
        return TaskCron(
            expression=cron_expression,
            spec=task.spec.model_copy(deep=True),
            config=task.config.model_copy(deep=True),
        )


# Overload for @task() or @task(retry=..., retry_delay=...)
@overload
def task(
    *,
    retry: float = 0,
    retry_delay: float | list[float] | None = None,
    middleware: list[Callable[..., Any]] | None = None
) -> Callable[[Callable[P, R]], Callable[P, Task]]:
    ...

# Overload for @task without parentheses
@overload
def task(func: Callable[P, R], /) -> Callable[P, Task]:
    ...

def task(
    func: Callable[P, R] | None = None,
    *,
    retry: float = 0,
    retry_delay: float | list[float] | None = None,
    middleware: list[Callable[..., Any]] | None = None
) -> Callable[[Callable[P, R]], Callable[P, Task]] | Callable[P, Task]:
    def decorator(func: Callable[P, R]) -> Callable[P, Task]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> Task:

            return TaskFactory.create_task(func, list(args), kwargs, retry, retry_delay, middleware)

        return wrapper

    # If called without parentheses (@task), func will be the decorated function
    if func is not None:
        return decorator(func)

    # If called with parentheses (@task() or @task(name="...")), return the decorator
    return decorator
